local function usage()
    print[[
soluadep: create a manifest of required packages

Usage: soluadep [opts] --static filename.lua
    Use static analysis on filename.lua to find module dependencies.
    Uses luac for source analysis, only finds literal require"mod".

Usage: soluadep [opts] scriptname [scriptargs]
    Run scriptname as usual; exit or call SOLUA_END() to print module
    dependencies found during run.

Creates solua.out, a file listing the dependencies and paths found.

Options:
    --binary "lfs lposix..."
        Declare modules to be binary-only (and not includable in a pure 
        script archive.) Suppresses warnings.
    --exclude "ml lom..."
        Do not include the named modules in output.
]]
os.exit(1)
end

-- We need to exclude what's already present. Unfortunately LUA_INIT
-- and this module itself use modules, so this is not quite the right thing:
-- 
-- local standalone_package_loaded = ml.update({},package.loaded)
--
local ml = require "ml"
--
-- Instead, just list things we do know about as of 2012.
local standalone_package_loaded = ml.set{"string", "debug", "package", "_G",
				 "io", "os", "table", "math", "coroutine"}
package.standalone_package_loaded = standalone_package_loaded
				      
-- If you're targeting another lua executable, add bit32 or whatever.

-- If you know you're using, say, LuaSocket and there's no chance supplying
-- it with an architecture-independent archive, arrange to place it here:

package.binonly_module = ml.split(os.getenv("SOLUA_BINONLY") or "")

-- package.binonly_module.socket=true

local explictly_excluded_modules = {}
local explictly_listed_modules = {}

local function warn(s)
    io.stderr:write("*-* "..s.."\n")
end

local function warn_binonly(mod)
    warn("Unable to find module |"..mod.."|; should it be in --binary or SOLUA_BINONLY?")
end

-- I'll use compat-5.1's search because it works better than mine. By the way,
-- is there some way of getting at the default path lua 5.2's ";;" component
-- specifies?

-- BEGIN compat-5.1
-- Copyright Kepler Project 2004-2006 (http://www.keplerproject.org/compat)
-- According to Lua 5.1
-- $Id: compat-5.1.lua,v 1.22 2006/02/20 21:12:47 carregal Exp $

local io,type = io,type
local find, format, gfind, gsub, sub = string.find, string.format, string.gfind, string.gsub, string.sub
local LUA_DIRSEP = '/'
local LUA_OFSEP = '_'
local LUA_PATH_MARK = '?'
local LUA_IGMARK = ':'

local function findfile (name, path)
    name = gsub (name, "%.", LUA_DIRSEP)
    for c in gfind (path, "[^;]+") do
	c = gsub (c, "%"..LUA_PATH_MARK, name)
	if ml.exists(c) then 
	    return c
	end
    end
    return nil -- not found
end

-- END compat-5.1

local string = string

-- Unfortunately you can't chain iterators directly like
--    file:lines():matching()
-- because io:lines() has multiple return values.
-- To avoid yet another cons, the iterator goes in varargs. Yuck
local function matchiter(pattern, ...)
    local iter = {...}
    return coroutine.wrap(
	function()
	    for line in unpack(iter) do 
		if not line then return nil end
		local rv = {string.match(line, pattern)}
		if rv[1] then
		    coroutine.yield(unpack(rv))
		end
	    end
	end)
end

local function matchlines(pattern, f)
    if f == nil then
	return matchiter(pattern, io.lines())
    else
	return matchiter(pattern, f:lines())
    end
end

--for number in matchlines("[0-9]+") do
--    print(number)
--end

local modulepaths = {}

-- ml.expand has a bug: if foo contains "$bar" then "${foo}" will be
-- double-substituted.

-- The static analysis is totally a rerun from lualint. Back once
-- again with the ill behavior.

local function popen_luac(sourcepath)
    assert(sourcepath)
    assert(ml.exists(sourcepath), "no file "..sourcepath)
    local badchar = sourcepath:match("[^%w :_/.-]+")
    if badchar then
	error(
	    ml.expand(
		'popen does not have any defined cross-platform quoting '..
		    'semantics; unknown character "$badchar" found '..
		    'in "$sourcepath".', 
		{badchar=badchar, sourcepath=sourcepath}))
    end
    
    return assert(io.popen('luac -p -l "'..sourcepath..'"'))
end

local function find_requires(sourcepath)
    -- This function is super-cautious because "luac -l" is not at all
    -- a documented interface, and blowing up on unexpected conditions
    -- is better than blithely plowing on.

    local f = popen_luac(sourcepath)

    assert(f:read() == "", "my luac returns a blank first line")
    assert(f:read():match("^main <"), "unanticipated first line")
    assert(f:read():match("^0%+ params,"), "main chunk takes parameters??")

    -- A typical line:
    -- 	11	[9]	GETTABLE 	4 2 -7	; "assert_arg"

    -- Match the mandatory part of each listing line up to the semicolon; 
    -- capture line, opcode, and any semicolon clause.
    -- If this fails to match, we don't understand anything about the listing.
    local pat = "^%s+[0-9]+%s+%[([0-9]+)]%s+([A-Z]+)%s+[-0-9%s]+(.*)"

    local requires = {}
    local state = "none"
    local candidate_module

    for line in f:lines() do
	if line:match("^%s*$") then
	    -- Blank line marks end of the chunk. Stop looking.
	    break
	end
	-- we're looking for a sequence of "GETGLOBAL require", a LOADK
	-- of the module name, and a CALL. Anybody who did local require=require
	-- has outsmarted us (and probably themselves too).
	local linenumber,opcode,tail = line:match(pat)
	assert(linenumber, "unknown assembly format")
	if state == "none" then
	    if opcode == "GETGLOBAL" and tail:match("^; require") then
		state = "found-require"
	    end
	elseif state == "found-require" then
	    -- 9	[8]	LOADK    	4 -6	; "pl.lexer"
	    module_name = tail:match('^; "(.*)"%s*$')
	    if opcode == "LOADK" and module_name then
		candidate_module = module_name
		state = "found-loadk"
	    else
		state = "none"
	    end
	elseif state == "found-loadk" then
	    if opcode == "CALL" then
		requires[#requires+1] = candidate_module
		--print(candidate_module)
	    end
	    state = "none"
	else error("bug in state machine") end
    end
    return requires
end

local function is_unloadable_module(mod)
    return package.binonly_module[mod] or package.standalone_package_loaded[mod] or explictly_excluded_modules[mod]
end

local function trace_static_dependencies_of_main(filename)
    local static_dependencies={}

    local trace_static_dependencies
    function trace_static_dependencies(rootmod, filename)
	static_dependencies[rootmod] = filename
	local found_requires = find_requires(filename)
	for _,mod in ipairs(found_requires) do
	    -- Skip if we've already registered it.
	    if not static_dependencies[mod] and not is_unloadable_module(mod) then
		local sub_filename = findfile(mod, package.path)
		if sub_filename then
		    trace_static_dependencies(mod, sub_filename)
		else
		    warn_binonly(mod)
		end
	    end
	end
    end
    trace_static_dependencies("_main", filename)
    return static_dependencies
end


-- And something to call at the end, if you like dynamic analysis:
local function find_dynamic_dependencies()
    local depends = {}
    for mod,_ in pairs(package.loaded) do
	if not is_unloadable_module(mod) then
	    local filename = findfile(mod, package.path)
	    if filename then
		depends[mod] = filename
	    else
		-- Well, we can't find where we loaded this module. This will
		-- be true for anything on cpath.
		warn_binonly(mod)
	    end
	end
    end
    return depends
end

-- sadly another approach, to deal with our own ml dependency
local real_require = require
package.explicit_require = {}
function require(mod)
    package.explicit_require[mod] = true
    return real_require(mod)
end

local function dump_deps(deps)
    ml.update(deps, explictly_excluded_modules)
    ml.update(deps, explictly_listed_modules)
    for k,_ in pairs(package.binonly_module) do
	deps[k] = 1010 -- binary, get it?
    end
    ml.tdump(deps)
    ml.tdump(deps, assert(io.open("solua.out", "w")))
end

-- In case we replace during script runs.
local real_os_exit=os.exit

local function SOLUA_END(n)
    -- let's not force ml down everyone's throat if they didn't ask for it
    if not package.explicit_require["ml"] then
	package.loaded.ml = nil
    end
    dump_deps(find_dynamic_dependencies())
    real_os_exit(n)
end

if arg then
    if #arg < 1 then
	usage()
    end

    local deps
    local scriptname
    local static = false

    local i = 0
    while i < #arg do
	i = i + 1
	if arg[i] == "--binary" then
	    i = i + 1
	    if not arg[i] then usage() end
	    ml.update(package.binonly_module, ml.set(ml.split(arg[i])))
	elseif arg[i] == "--exclude" then 
	    i = i + 1
	    if not arg[i] then usage() end
	    ml.update(explictly_excluded_modules, ml.set(ml.split(arg[i])))
	elseif arg[i] == "--static" then
	    static = true
	else
	    scriptname = arg[i]
	    break
	end
    end
    if not scriptname then usage() end
    arg = ml.sub(arg, i+1)
    if static then
	deps = trace_static_dependencies_of_main(scriptname)
	dump_deps(deps)
    else
	explictly_listed_modules._main = scriptname
	os.exit = SOLUA_END
	dofile(scriptname)
	-- Oh, I'll do it for you.
	os.exit(0)
    end
end
